# coding: utf-8

"""
    Argo Server API

    You can get examples of requests and responses by using the CLI with `--gloglevel=9`, e.g. `argo list --gloglevel=9`  # noqa: E501

    The version of the OpenAPI document: v2.12.2
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from argo.workflows.client.configuration import Configuration


class V1alpha1Template(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'affinity': 'V1Affinity',
        'archive_location': 'V1alpha1ArtifactLocation',
        'arguments': 'V1alpha1Arguments',
        'automount_service_account_token': 'bool',
        'container': 'V1Container',
        'daemon': 'bool',
        'dag': 'V1alpha1DAGTemplate',
        'executor': 'V1alpha1ExecutorConfig',
        'host_aliases': 'list[V1HostAlias]',
        'http': 'dict',
        'init_containers': 'list[V1alpha1UserContainer]',
        'inputs': 'V1alpha1Inputs',
        'memoize': 'V1alpha1Memoize',
        'metadata': 'V1alpha1Metadata',
        'metrics': 'V1alpha1Metrics',
        'name': 'str',
        'node_selector': 'dict(str, str)',
        'outputs': 'V1alpha1Outputs',
        'parallelism': 'int',
        'pod_spec_patch': 'str',
        'priority': 'int',
        'priority_class_name': 'str',
        'resource': 'V1alpha1ResourceTemplate',
        'retry_strategy': 'V1alpha1RetryStrategy',
        'scheduler_name': 'str',
        'script': 'V1alpha1ScriptTemplate',
        'security_context': 'V1PodSecurityContext',
        'service_account_name': 'str',
        'sidecars': 'list[V1alpha1UserContainer]',
        'steps': 'list[list[V1alpha1WorkflowStep]]',
        'suspend': 'V1alpha1SuspendTemplate',
        'synchronization': 'V1alpha1Synchronization',
        'template': 'str',
        'template_ref': 'V1alpha1TemplateRef',
        'timeout': 'str',
        'tolerations': 'list[V1Toleration]',
        'volumes': 'list[V1Volume]'
    }

    attribute_map = {
        'affinity': 'affinity',
        'archive_location': 'archiveLocation',
        'arguments': 'arguments',
        'automount_service_account_token': 'automountServiceAccountToken',
        'container': 'container',
        'daemon': 'daemon',
        'dag': 'dag',
        'executor': 'executor',
        'host_aliases': 'hostAliases',
        'http': 'http',
        'init_containers': 'initContainers',
        'inputs': 'inputs',
        'memoize': 'memoize',
        'metadata': 'metadata',
        'metrics': 'metrics',
        'name': 'name',
        'node_selector': 'nodeSelector',
        'outputs': 'outputs',
        'parallelism': 'parallelism',
        'pod_spec_patch': 'podSpecPatch',
        'priority': 'priority',
        'priority_class_name': 'priorityClassName',
        'resource': 'resource',
        'retry_strategy': 'retryStrategy',
        'scheduler_name': 'schedulerName',
        'script': 'script',
        'security_context': 'securityContext',
        'service_account_name': 'serviceAccountName',
        'sidecars': 'sidecars',
        'steps': 'steps',
        'suspend': 'suspend',
        'synchronization': 'synchronization',
        'template': 'template',
        'template_ref': 'templateRef',
        'timeout': 'timeout',
        'tolerations': 'tolerations',
        'volumes': 'volumes'
    }

    def __init__(self, affinity=None, archive_location=None, arguments=None, automount_service_account_token=None, container=None, daemon=None, dag=None, executor=None, host_aliases=None, http=None, init_containers=None, inputs=None, memoize=None, metadata=None, metrics=None, name=None, node_selector=None, outputs=None, parallelism=None, pod_spec_patch=None, priority=None, priority_class_name=None, resource=None, retry_strategy=None, scheduler_name=None, script=None, security_context=None, service_account_name=None, sidecars=None, steps=None, suspend=None, synchronization=None, template=None, template_ref=None, timeout=None, tolerations=None, volumes=None, local_vars_configuration=None):  # noqa: E501
        """V1alpha1Template - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._affinity = None
        self._archive_location = None
        self._arguments = None
        self._automount_service_account_token = None
        self._container = None
        self._daemon = None
        self._dag = None
        self._executor = None
        self._host_aliases = None
        self._http = None
        self._init_containers = None
        self._inputs = None
        self._memoize = None
        self._metadata = None
        self._metrics = None
        self._name = None
        self._node_selector = None
        self._outputs = None
        self._parallelism = None
        self._pod_spec_patch = None
        self._priority = None
        self._priority_class_name = None
        self._resource = None
        self._retry_strategy = None
        self._scheduler_name = None
        self._script = None
        self._security_context = None
        self._service_account_name = None
        self._sidecars = None
        self._steps = None
        self._suspend = None
        self._synchronization = None
        self._template = None
        self._template_ref = None
        self._timeout = None
        self._tolerations = None
        self._volumes = None
        self.discriminator = None

        if affinity is not None:
            self.affinity = affinity
        if archive_location is not None:
            self.archive_location = archive_location
        if arguments is not None:
            self.arguments = arguments
        if automount_service_account_token is not None:
            self.automount_service_account_token = automount_service_account_token  # noqa: E501
        if container is not None:
            self.container = container
        if daemon is not None:
            self.daemon = daemon
        if dag is not None:
            self.dag = dag
        if executor is not None:
            self.executor = executor
        if host_aliases is not None:
            self.host_aliases = host_aliases
        if http is not None:
            self.http = http
        if init_containers is not None:
            self.init_containers = init_containers
        if inputs is not None:
            self.inputs = inputs
        if memoize is not None:
            self.memoize = memoize
        if metadata is not None:
            self.metadata = metadata
        if metrics is not None:
            self.metrics = metrics
        self.name = name
        if node_selector is not None:
            self.node_selector = node_selector
        if outputs is not None:
            self.outputs = outputs
        if parallelism is not None:
            self.parallelism = parallelism
        if pod_spec_patch is not None:
            self.pod_spec_patch = pod_spec_patch
        if priority is not None:
            self.priority = priority
        if priority_class_name is not None:
            self.priority_class_name = priority_class_name
        if resource is not None:
            self.resource = resource
        if retry_strategy is not None:
            self.retry_strategy = retry_strategy
        if scheduler_name is not None:
            self.scheduler_name = scheduler_name
        if script is not None:
            self.script = script
        if security_context is not None:
            self.security_context = security_context
        if service_account_name is not None:
            self.service_account_name = service_account_name
        if sidecars is not None:
            self.sidecars = sidecars
        if steps is not None:
            self.steps = steps
        if suspend is not None:
            self.suspend = suspend
        if synchronization is not None:
            self.synchronization = synchronization
        if template is not None:
            self.template = template
        if template_ref is not None:
            self.template_ref = template_ref
        if timeout is not None:
            self.timeout = timeout
        if tolerations is not None:
            self.tolerations = tolerations
        if volumes is not None:
            self.volumes = volumes

    @property
    def affinity(self):
        """Gets the affinity of this V1alpha1Template.  # noqa: E501


        :return: The affinity of this V1alpha1Template.  # noqa: E501
        :rtype: V1Affinity
        """
        return self._affinity

    @affinity.setter
    def affinity(self, affinity):
        """Sets the affinity of this V1alpha1Template.


        :param affinity: The affinity of this V1alpha1Template.  # noqa: E501
        :type: V1Affinity
        """

        self._affinity = affinity

    @property
    def archive_location(self):
        """Gets the archive_location of this V1alpha1Template.  # noqa: E501


        :return: The archive_location of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1ArtifactLocation
        """
        return self._archive_location

    @archive_location.setter
    def archive_location(self, archive_location):
        """Sets the archive_location of this V1alpha1Template.


        :param archive_location: The archive_location of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1ArtifactLocation
        """

        self._archive_location = archive_location

    @property
    def arguments(self):
        """Gets the arguments of this V1alpha1Template.  # noqa: E501


        :return: The arguments of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1Arguments
        """
        return self._arguments

    @arguments.setter
    def arguments(self, arguments):
        """Sets the arguments of this V1alpha1Template.


        :param arguments: The arguments of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1Arguments
        """

        self._arguments = arguments

    @property
    def automount_service_account_token(self):
        """Gets the automount_service_account_token of this V1alpha1Template.  # noqa: E501

        AutomountServiceAccountToken indicates whether a service account token should be automatically mounted in pods. ServiceAccountName of ExecutorConfig must be specified if this value is false.  # noqa: E501

        :return: The automount_service_account_token of this V1alpha1Template.  # noqa: E501
        :rtype: bool
        """
        return self._automount_service_account_token

    @automount_service_account_token.setter
    def automount_service_account_token(self, automount_service_account_token):
        """Sets the automount_service_account_token of this V1alpha1Template.

        AutomountServiceAccountToken indicates whether a service account token should be automatically mounted in pods. ServiceAccountName of ExecutorConfig must be specified if this value is false.  # noqa: E501

        :param automount_service_account_token: The automount_service_account_token of this V1alpha1Template.  # noqa: E501
        :type: bool
        """

        self._automount_service_account_token = automount_service_account_token

    @property
    def container(self):
        """Gets the container of this V1alpha1Template.  # noqa: E501


        :return: The container of this V1alpha1Template.  # noqa: E501
        :rtype: V1Container
        """
        return self._container

    @container.setter
    def container(self, container):
        """Sets the container of this V1alpha1Template.


        :param container: The container of this V1alpha1Template.  # noqa: E501
        :type: V1Container
        """

        self._container = container

    @property
    def daemon(self):
        """Gets the daemon of this V1alpha1Template.  # noqa: E501

        Deamon will allow a workflow to proceed to the next step so long as the container reaches readiness  # noqa: E501

        :return: The daemon of this V1alpha1Template.  # noqa: E501
        :rtype: bool
        """
        return self._daemon

    @daemon.setter
    def daemon(self, daemon):
        """Sets the daemon of this V1alpha1Template.

        Deamon will allow a workflow to proceed to the next step so long as the container reaches readiness  # noqa: E501

        :param daemon: The daemon of this V1alpha1Template.  # noqa: E501
        :type: bool
        """

        self._daemon = daemon

    @property
    def dag(self):
        """Gets the dag of this V1alpha1Template.  # noqa: E501


        :return: The dag of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1DAGTemplate
        """
        return self._dag

    @dag.setter
    def dag(self, dag):
        """Sets the dag of this V1alpha1Template.


        :param dag: The dag of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1DAGTemplate
        """

        self._dag = dag

    @property
    def executor(self):
        """Gets the executor of this V1alpha1Template.  # noqa: E501


        :return: The executor of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1ExecutorConfig
        """
        return self._executor

    @executor.setter
    def executor(self, executor):
        """Sets the executor of this V1alpha1Template.


        :param executor: The executor of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1ExecutorConfig
        """

        self._executor = executor

    @property
    def host_aliases(self):
        """Gets the host_aliases of this V1alpha1Template.  # noqa: E501

        HostAliases is an optional list of hosts and IPs that will be injected into the pod spec  # noqa: E501

        :return: The host_aliases of this V1alpha1Template.  # noqa: E501
        :rtype: list[V1HostAlias]
        """
        return self._host_aliases

    @host_aliases.setter
    def host_aliases(self, host_aliases):
        """Sets the host_aliases of this V1alpha1Template.

        HostAliases is an optional list of hosts and IPs that will be injected into the pod spec  # noqa: E501

        :param host_aliases: The host_aliases of this V1alpha1Template.  # noqa: E501
        :type: list[V1HostAlias]
        """

        self._host_aliases = host_aliases

    @property
    def http(self):
        """Gets the http of this V1alpha1Template.  # noqa: E501

        :return: The http of this V1alpha1Template.  # noqa: E501
        :rtype: dict
        """
        return self._http

    @http.setter
    def http(self, http):
        """Sets the http of this V1alpha1Template.

        :param http: The http of this V1alpha1Template.  # noqa: E501
        :type: dict
        """

        self._http = http

    @property
    def init_containers(self):
        """Gets the init_containers of this V1alpha1Template.  # noqa: E501

        InitContainers is a list of containers which run before the main container.  # noqa: E501

        :return: The init_containers of this V1alpha1Template.  # noqa: E501
        :rtype: list[V1alpha1UserContainer]
        """
        return self._init_containers

    @init_containers.setter
    def init_containers(self, init_containers):
        """Sets the init_containers of this V1alpha1Template.

        InitContainers is a list of containers which run before the main container.  # noqa: E501

        :param init_containers: The init_containers of this V1alpha1Template.  # noqa: E501
        :type: list[V1alpha1UserContainer]
        """

        self._init_containers = init_containers

    @property
    def inputs(self):
        """Gets the inputs of this V1alpha1Template.  # noqa: E501


        :return: The inputs of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1Inputs
        """
        return self._inputs

    @inputs.setter
    def inputs(self, inputs):
        """Sets the inputs of this V1alpha1Template.


        :param inputs: The inputs of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1Inputs
        """

        self._inputs = inputs

    @property
    def memoize(self):
        """Gets the memoize of this V1alpha1Template.  # noqa: E501


        :return: The memoize of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1Memoize
        """
        return self._memoize

    @memoize.setter
    def memoize(self, memoize):
        """Sets the memoize of this V1alpha1Template.


        :param memoize: The memoize of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1Memoize
        """

        self._memoize = memoize

    @property
    def metadata(self):
        """Gets the metadata of this V1alpha1Template.  # noqa: E501


        :return: The metadata of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1Metadata
        """
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        """Sets the metadata of this V1alpha1Template.


        :param metadata: The metadata of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1Metadata
        """

        self._metadata = metadata

    @property
    def metrics(self):
        """Gets the metrics of this V1alpha1Template.  # noqa: E501


        :return: The metrics of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1Metrics
        """
        return self._metrics

    @metrics.setter
    def metrics(self, metrics):
        """Sets the metrics of this V1alpha1Template.


        :param metrics: The metrics of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1Metrics
        """

        self._metrics = metrics

    @property
    def name(self):
        """Gets the name of this V1alpha1Template.  # noqa: E501

        Name is the name of the template  # noqa: E501

        :return: The name of this V1alpha1Template.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this V1alpha1Template.

        Name is the name of the template  # noqa: E501

        :param name: The name of this V1alpha1Template.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def node_selector(self):
        """Gets the node_selector of this V1alpha1Template.  # noqa: E501

        NodeSelector is a selector to schedule this step of the workflow to be run on the selected node(s). Overrides the selector set at the workflow level.  # noqa: E501

        :return: The node_selector of this V1alpha1Template.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._node_selector

    @node_selector.setter
    def node_selector(self, node_selector):
        """Sets the node_selector of this V1alpha1Template.

        NodeSelector is a selector to schedule this step of the workflow to be run on the selected node(s). Overrides the selector set at the workflow level.  # noqa: E501

        :param node_selector: The node_selector of this V1alpha1Template.  # noqa: E501
        :type: dict(str, str)
        """

        self._node_selector = node_selector

    @property
    def outputs(self):
        """Gets the outputs of this V1alpha1Template.  # noqa: E501


        :return: The outputs of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1Outputs
        """
        return self._outputs

    @outputs.setter
    def outputs(self, outputs):
        """Sets the outputs of this V1alpha1Template.


        :param outputs: The outputs of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1Outputs
        """

        self._outputs = outputs

    @property
    def parallelism(self):
        """Gets the parallelism of this V1alpha1Template.  # noqa: E501

        Parallelism limits the max total parallel pods that can execute at the same time within the boundaries of this template invocation. If additional steps/dag templates are invoked, the pods created by those templates will not be counted towards this total.  # noqa: E501

        :return: The parallelism of this V1alpha1Template.  # noqa: E501
        :rtype: int
        """
        return self._parallelism

    @parallelism.setter
    def parallelism(self, parallelism):
        """Sets the parallelism of this V1alpha1Template.

        Parallelism limits the max total parallel pods that can execute at the same time within the boundaries of this template invocation. If additional steps/dag templates are invoked, the pods created by those templates will not be counted towards this total.  # noqa: E501

        :param parallelism: The parallelism of this V1alpha1Template.  # noqa: E501
        :type: int
        """

        self._parallelism = parallelism

    @property
    def pod_spec_patch(self):
        """Gets the pod_spec_patch of this V1alpha1Template.  # noqa: E501

        PodSpecPatch holds strategic merge patch to apply against the pod spec. Allows parameterization of container fields which are not strings (e.g. resource limits).  # noqa: E501

        :return: The pod_spec_patch of this V1alpha1Template.  # noqa: E501
        :rtype: str
        """
        return self._pod_spec_patch

    @pod_spec_patch.setter
    def pod_spec_patch(self, pod_spec_patch):
        """Sets the pod_spec_patch of this V1alpha1Template.

        PodSpecPatch holds strategic merge patch to apply against the pod spec. Allows parameterization of container fields which are not strings (e.g. resource limits).  # noqa: E501

        :param pod_spec_patch: The pod_spec_patch of this V1alpha1Template.  # noqa: E501
        :type: str
        """

        self._pod_spec_patch = pod_spec_patch

    @property
    def priority(self):
        """Gets the priority of this V1alpha1Template.  # noqa: E501

        Priority to apply to workflow pods.  # noqa: E501

        :return: The priority of this V1alpha1Template.  # noqa: E501
        :rtype: int
        """
        return self._priority

    @priority.setter
    def priority(self, priority):
        """Sets the priority of this V1alpha1Template.

        Priority to apply to workflow pods.  # noqa: E501

        :param priority: The priority of this V1alpha1Template.  # noqa: E501
        :type: int
        """

        self._priority = priority

    @property
    def priority_class_name(self):
        """Gets the priority_class_name of this V1alpha1Template.  # noqa: E501

        PriorityClassName to apply to workflow pods.  # noqa: E501

        :return: The priority_class_name of this V1alpha1Template.  # noqa: E501
        :rtype: str
        """
        return self._priority_class_name

    @priority_class_name.setter
    def priority_class_name(self, priority_class_name):
        """Sets the priority_class_name of this V1alpha1Template.

        PriorityClassName to apply to workflow pods.  # noqa: E501

        :param priority_class_name: The priority_class_name of this V1alpha1Template.  # noqa: E501
        :type: str
        """

        self._priority_class_name = priority_class_name

    @property
    def resource(self):
        """Gets the resource of this V1alpha1Template.  # noqa: E501


        :return: The resource of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1ResourceTemplate
        """
        return self._resource

    @resource.setter
    def resource(self, resource):
        """Sets the resource of this V1alpha1Template.


        :param resource: The resource of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1ResourceTemplate
        """

        self._resource = resource

    @property
    def retry_strategy(self):
        """Gets the retry_strategy of this V1alpha1Template.  # noqa: E501


        :return: The retry_strategy of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1RetryStrategy
        """
        return self._retry_strategy

    @retry_strategy.setter
    def retry_strategy(self, retry_strategy):
        """Sets the retry_strategy of this V1alpha1Template.


        :param retry_strategy: The retry_strategy of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1RetryStrategy
        """

        self._retry_strategy = retry_strategy

    @property
    def scheduler_name(self):
        """Gets the scheduler_name of this V1alpha1Template.  # noqa: E501

        If specified, the pod will be dispatched by specified scheduler. Or it will be dispatched by workflow scope scheduler if specified. If neither specified, the pod will be dispatched by default scheduler.  # noqa: E501

        :return: The scheduler_name of this V1alpha1Template.  # noqa: E501
        :rtype: str
        """
        return self._scheduler_name

    @scheduler_name.setter
    def scheduler_name(self, scheduler_name):
        """Sets the scheduler_name of this V1alpha1Template.

        If specified, the pod will be dispatched by specified scheduler. Or it will be dispatched by workflow scope scheduler if specified. If neither specified, the pod will be dispatched by default scheduler.  # noqa: E501

        :param scheduler_name: The scheduler_name of this V1alpha1Template.  # noqa: E501
        :type: str
        """

        self._scheduler_name = scheduler_name

    @property
    def script(self):
        """Gets the script of this V1alpha1Template.  # noqa: E501


        :return: The script of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1ScriptTemplate
        """
        return self._script

    @script.setter
    def script(self, script):
        """Sets the script of this V1alpha1Template.


        :param script: The script of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1ScriptTemplate
        """

        self._script = script

    @property
    def security_context(self):
        """Gets the security_context of this V1alpha1Template.  # noqa: E501


        :return: The security_context of this V1alpha1Template.  # noqa: E501
        :rtype: V1PodSecurityContext
        """
        return self._security_context

    @security_context.setter
    def security_context(self, security_context):
        """Sets the security_context of this V1alpha1Template.


        :param security_context: The security_context of this V1alpha1Template.  # noqa: E501
        :type: V1PodSecurityContext
        """

        self._security_context = security_context

    @property
    def service_account_name(self):
        """Gets the service_account_name of this V1alpha1Template.  # noqa: E501

        ServiceAccountName to apply to workflow pods  # noqa: E501

        :return: The service_account_name of this V1alpha1Template.  # noqa: E501
        :rtype: str
        """
        return self._service_account_name

    @service_account_name.setter
    def service_account_name(self, service_account_name):
        """Sets the service_account_name of this V1alpha1Template.

        ServiceAccountName to apply to workflow pods  # noqa: E501

        :param service_account_name: The service_account_name of this V1alpha1Template.  # noqa: E501
        :type: str
        """

        self._service_account_name = service_account_name

    @property
    def sidecars(self):
        """Gets the sidecars of this V1alpha1Template.  # noqa: E501

        Sidecars is a list of containers which run alongside the main container Sidecars are automatically killed when the main container completes  # noqa: E501

        :return: The sidecars of this V1alpha1Template.  # noqa: E501
        :rtype: list[V1alpha1UserContainer]
        """
        return self._sidecars

    @sidecars.setter
    def sidecars(self, sidecars):
        """Sets the sidecars of this V1alpha1Template.

        Sidecars is a list of containers which run alongside the main container Sidecars are automatically killed when the main container completes  # noqa: E501

        :param sidecars: The sidecars of this V1alpha1Template.  # noqa: E501
        :type: list[V1alpha1UserContainer]
        """

        self._sidecars = sidecars

    @property
    def steps(self):
        """Gets the steps of this V1alpha1Template.  # noqa: E501

        Steps define a series of sequential/parallel workflow steps  # noqa: E501

        :return: The steps of this V1alpha1Template.  # noqa: E501
        :rtype: list[list[V1alpha1WorkflowStep]]
        """
        return self._steps

    @steps.setter
    def steps(self, steps):
        """Sets the steps of this V1alpha1Template.

        Steps define a series of sequential/parallel workflow steps  # noqa: E501

        :param steps: The steps of this V1alpha1Template.  # noqa: E501
        :type: list[list[V1alpha1WorkflowStep]]
        """

        self._steps = steps

    @property
    def suspend(self):
        """Gets the suspend of this V1alpha1Template.  # noqa: E501


        :return: The suspend of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1SuspendTemplate
        """
        return self._suspend

    @suspend.setter
    def suspend(self, suspend):
        """Sets the suspend of this V1alpha1Template.


        :param suspend: The suspend of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1SuspendTemplate
        """

        self._suspend = suspend

    @property
    def synchronization(self):
        """Gets the synchronization of this V1alpha1Template.  # noqa: E501


        :return: The synchronization of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1Synchronization
        """
        return self._synchronization

    @synchronization.setter
    def synchronization(self, synchronization):
        """Sets the synchronization of this V1alpha1Template.


        :param synchronization: The synchronization of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1Synchronization
        """

        self._synchronization = synchronization

    @property
    def template(self):
        """Gets the template of this V1alpha1Template.  # noqa: E501

        Template is the name of the template which is used as the base of this template. DEPRECATED: This field is not used.  # noqa: E501

        :return: The template of this V1alpha1Template.  # noqa: E501
        :rtype: str
        """
        return self._template

    @template.setter
    def template(self, template):
        """Sets the template of this V1alpha1Template.

        Template is the name of the template which is used as the base of this template. DEPRECATED: This field is not used.  # noqa: E501

        :param template: The template of this V1alpha1Template.  # noqa: E501
        :type: str
        """

        self._template = template

    @property
    def template_ref(self):
        """Gets the template_ref of this V1alpha1Template.  # noqa: E501


        :return: The template_ref of this V1alpha1Template.  # noqa: E501
        :rtype: V1alpha1TemplateRef
        """
        return self._template_ref

    @template_ref.setter
    def template_ref(self, template_ref):
        """Sets the template_ref of this V1alpha1Template.


        :param template_ref: The template_ref of this V1alpha1Template.  # noqa: E501
        :type: V1alpha1TemplateRef
        """

        self._template_ref = template_ref

    @property
    def timeout(self):
        """Gets the timeout of this V1alpha1Template.  # noqa: E501

        Timout allows to set the total node execution timeout duration counting from the node's start time. This duration also includes time in which the node spends in Pending state. This duration may not be applied to Step or DAG templates.  # noqa: E501

        :return: The timeout of this V1alpha1Template.  # noqa: E501
        :rtype: str
        """
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        """Sets the timeout of this V1alpha1Template.

        Timout allows to set the total node execution timeout duration counting from the node's start time. This duration also includes time in which the node spends in Pending state. This duration may not be applied to Step or DAG templates.  # noqa: E501

        :param timeout: The timeout of this V1alpha1Template.  # noqa: E501
        :type: str
        """

        self._timeout = timeout

    @property
    def tolerations(self):
        """Gets the tolerations of this V1alpha1Template.  # noqa: E501

        Tolerations to apply to workflow pods.  # noqa: E501

        :return: The tolerations of this V1alpha1Template.  # noqa: E501
        :rtype: list[V1Toleration]
        """
        return self._tolerations

    @tolerations.setter
    def tolerations(self, tolerations):
        """Sets the tolerations of this V1alpha1Template.

        Tolerations to apply to workflow pods.  # noqa: E501

        :param tolerations: The tolerations of this V1alpha1Template.  # noqa: E501
        :type: list[V1Toleration]
        """

        self._tolerations = tolerations

    @property
    def volumes(self):
        """Gets the volumes of this V1alpha1Template.  # noqa: E501

        Volumes is a list of volumes that can be mounted by containers in a template.  # noqa: E501

        :return: The volumes of this V1alpha1Template.  # noqa: E501
        :rtype: list[V1Volume]
        """
        return self._volumes

    @volumes.setter
    def volumes(self, volumes):
        """Sets the volumes of this V1alpha1Template.

        Volumes is a list of volumes that can be mounted by containers in a template.  # noqa: E501

        :param volumes: The volumes of this V1alpha1Template.  # noqa: E501
        :type: list[V1Volume]
        """

        self._volumes = volumes

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, V1alpha1Template):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1alpha1Template):
            return True

        return self.to_dict() != other.to_dict()



import json
from copy import deepcopy
from typing import Dict, List, Optional, Union

from .common import field_errmsg, field_regex
from .config import config as global_config
from .config import s3_config
from .io import PVC, InputParameter, Inputs, OutputParameter, Outputs
from .utils import randstr

try:
    from argo.workflows.client import (V1Affinity, V1alpha1Cache,
                                       V1alpha1Memoize, V1alpha1Metadata,
                                       V1alpha1ResourceTemplate,
                                       V1alpha1ScriptTemplate,
                                       V1alpha1UserContainer,
                                       V1ConfigMapKeySelector, V1EnvVar,
                                       V1EnvVarSource, V1ResourceRequirements,
                                       V1SecretKeySelector, V1Toleration,
                                       V1Volume, V1VolumeMount)
    from argo.workflows.client.configuration import Configuration

    import kubernetes

    from .client import (V1alpha1RetryStrategy, V1alpha1Template)
except Exception:
    V1Affinity = object
    V1alpha1ResourceTemplate = object
    V1alpha1RetryStrategy = object
    V1alpha1UserContainer = object
    V1EnvVarSource = object
    V1Toleration = object
    V1Volume = object
    V1VolumeMount = object


class Secret:
    def __init__(self, value=None, name=None, key="secret"):
        self.secret_name = "dflow-" + randstr(15) if name is None else name
        self.secret_key = key
        self.value = value
        if value is not None:
            data = {self.secret_key: value}
            secret = kubernetes.client.V1Secret(
                string_data=data,
                metadata=kubernetes.client.V1ObjectMeta(name=self.secret_name),
                type="Opaque")
            with get_k8s_client() as k8s_client:
                core_v1_api = kubernetes.client.CoreV1Api(k8s_client)
                core_v1_api.api_client.call_api(
                    '/api/v1/namespaces/%s/secrets' %
                    global_config["namespace"],
                    'POST', body=secret, response_type='V1Secret',
                    header_params=global_config["http_headers"],
                    _return_http_data_only=True)

    def __repr__(self):
        return "Secret(name='%s', key='%s')" % (self.secret_name,
                                                self.secret_key)


def get_k8s_client(k8s_api_server=None, token=None, k8s_config_file=None):
    if k8s_api_server is None:
        k8s_api_server = global_config["k8s_api_server"]
    if token is None:
        token = global_config["token"]
    if k8s_config_file is None:
        k8s_config_file = global_config["k8s_config_file"]
    if k8s_api_server is not None:
        k8s_configuration = kubernetes.client.Configuration(
            host=k8s_api_server)
        k8s_configuration.verify_ssl = False
        if token is None:
            k8s_client = kubernetes.client.ApiClient(k8s_configuration)
        else:
            k8s_client = kubernetes.client.ApiClient(
                k8s_configuration, header_name='Authorization',
                header_value='Bearer %s' % token)
        return k8s_client
    else:
        return kubernetes.config.new_client_from_config(
            config_file=k8s_config_file)


class OPTemplate:
    def __init__(
            self,
            name: Optional[str] = None,
            inputs: Optional[Inputs] = None,
            outputs: Optional[Outputs] = None,
            memoize_key: Optional[str] = None,
            pvcs: Optional[List[PVC]] = None,
            annotations: Dict[str, str] = None,
            labels: Dict[str, str] = None,
    ) -> None:
        if name is None:
            name = randstr()
        # force lowercase to fix RFC 1123
        name = name.lower()
        assert field_regex.match(name), "Invalid OP template name '%s': %s" % (
            name, field_errmsg)
        self.name = name
        if inputs is not None:
            self.inputs = inputs
        else:
            self.inputs = Inputs(template=self)
        if outputs is not None:
            self.outputs = outputs
        else:
            self.outputs = Outputs(template=self)
        self.memoize_key = memoize_key
        self.memoize = None
        if pvcs is None:
            pvcs = []
        self.pvcs = pvcs
        if annotations is None:
            annotations = {}
        self.annotations = annotations
        self.labels = labels if labels is not None else {}
        self.modified = False

    @classmethod
    def from_dict(cls, d):
        kwargs = {
            "name": d.get("name", None),
            "inputs": Inputs.from_dict(d.get("inputs", {})),
            "outputs": Outputs.from_dict(d.get("outputs", {})),
            "memoize_key": d.get("memoize", {}).get("key", None),
            "annotations": d.get("metadata", {}).get("annotations", None),
        }
        return cls(**kwargs)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key == "inputs":
            assert isinstance(value, Inputs)
            value.set_template(self)
        elif key == "outputs":
            assert isinstance(value, Outputs)
            value.set_template(self)

    def handle_key(self, memoize_prefix=None, memoize_configmap="dflow"):
        if "dflow_key" in self.inputs.parameters:
            if memoize_prefix is not None:
                self.memoize_key = "%s-{{inputs.parameters.dflow_key}}" \
                    % memoize_prefix

            if global_config["save_keys_in_global_outputs"]:
                if hasattr(self, "image"):
                    self.outputs.parameters["dflow_global"] = OutputParameter(
                        value="{{pod.name}}", global_name="dflow_key_"
                        "{{inputs.parameters.dflow_key}}",
                    )
                else:
                    self.outputs.parameters["dflow_global"] = OutputParameter(
                        value_from_parameter="non-exists",
                        default="{{node.id}}", global_name="dflow_key_"
                        "{{inputs.parameters.dflow_key}}",
                    )
        elif "dflow_group_key" in self.inputs.parameters:
            if memoize_prefix is not None:
                self.memoize_key = \
                    "%s-{{inputs.parameters.dflow_group_key}}-init-artifact" \
                    % memoize_prefix

        if self.memoize_key is not None:
            # Is it a bug of Argo?
            if self.memoize_key.find("workflow.name") != -1:
                self.inputs.parameters["workflow_name"] = InputParameter(
                    value="{{workflow.name}}")
                self.memoize_key = self.memoize_key.replace(
                    "workflow.name", "inputs.parameters.workflow_name")
            config = Configuration()
            config.client_side_validation = False
            self.memoize = V1alpha1Memoize(
                key=self.memoize_key,
                local_vars_configuration=config, cache=V1alpha1Cache(
                    config_map=V1ConfigMapKeySelector(
                        name="%s-%s" % (memoize_configmap, self.memoize_key),
                        local_vars_configuration=config)))

    def copy(self):
        if self.modified:
            return self
        new_template = deepcopy(self)
        new_template.name += "-" + randstr()
        new_template.modified = True
        return new_template

    def deepcopy(self):
        if self.modified:
            return self
        new_template = deepcopy(self)
        new_template.name += "-" + randstr()
        new_template.modified = True
        return new_template


class HTTPOPTemplate(OPTemplate):
    def __init__(
            self,
            url: str,
            name: Optional[str] = None,
            inputs: Optional[Inputs] = None,
            outputs: Optional[Outputs] = None,
            memoize_key: Optional[str] = None,
            annotations: Dict[str, str] = None,
            labels: Dict[str, str] = None,
            method: str = "get",
            headers: Dict[str, str] = None,
            timeout: Optional[int] = None,
            success_condition: Optional[str] = None,
            body: Union[str, bytes] = None,
            insecure_skip_verify: bool = False,
    ) -> None:
        super().__init__(name=name, inputs=inputs, outputs=outputs,
                         memoize_key=memoize_key, annotations=annotations,
                         labels=labels)
        self.url = url
        self.method = method
        if headers is None:
            headers = {}
        self.headers = headers
        self.timeout = timeout
        self.success_condition = success_condition
        self.body = body
        self.insecure_skip_verify = insecure_skip_verify

    def convert_to_argo(self, memoize_prefix=None, memoize_configmap="dflow"):
        self.handle_key(memoize_prefix, memoize_configmap)
        return V1alpha1Template(
            name=self.name,
            metadata=V1alpha1Metadata(
                annotations=self.annotations),
            inputs=self.inputs.convert_to_argo(),
            outputs=self.outputs.convert_to_argo(),
            memoize=self.memoize,
            http=dict(
                method=self.method,
                url=self.url,
                headers=[{"name": k, "value": v}
                         for k, v in self.headers.items()],
                timeoutSeconds=self.timeout,
                successCondition=self.success_condition,
                body=self.body if isinstance(self.body, str) else None,
                bodyFrom={"bytes": self.body} if isinstance(self.body, bytes)
                else None,
                insecureSkipVerify=self.insecure_skip_verify,
            )
        )


class ScriptOPTemplate(OPTemplate):
    """
    Script OP template

    Args:
        name: the name of the OP template
        inputs: input parameters and input artifacts
        outputs: output parameters and output artifacts
        image: image the template uses
        command: command to run the script
        script: script
        volumes: volumes the template uses
        mounts: volumes the template mounts
        init_progress: a str representing the initial progress
        timeout: timeout of the OP template
        retry_strategy: retry strategy of the OP template
        memoize_key: memoized key of the OP template
        pvcs: PVCs need to be declared
        image_pull_policy: Always, IfNotPresent, Never
        annotations: annotations for the OP template
        labels: labels for the OP template
        node_selector: node selector when scheduling the pod
        tolerations: tolerations of taints when scheduling the pod
        affinity: affinity when scheduling the pod
        requests: a dict of resource requests
        limits: a dict of resource limits
        envs: environment variables
    """

    def __init__(
            self,
            name: Optional[str] = None,
            inputs: Optional[Inputs] = None,
            outputs: Optional[Outputs] = None,
            memoize_key: Optional[str] = None,
            pvcs: Optional[List[PVC]] = None,
            annotations: Dict[str, str] = None,
            labels: Dict[str, str] = None,
            node_selector: Dict[str, str] = None,
            tolerations: List[V1Toleration] = None,
            affinity: V1Affinity = None,
            image: Optional[str] = None,
            command: Union[str, List[str]] = None,
            script: Optional[str] = None,
            volumes: Optional[List[V1Volume]] = None,
            mounts: Optional[List[V1VolumeMount]] = None,
            init_progress: str = "0/1",
            timeout: Optional[str] = None,
            retry_strategy: Optional[V1alpha1RetryStrategy] = None,
            resource: Optional[V1alpha1ResourceTemplate] = None,
            image_pull_policy: Optional[str] = None,
            requests: Dict[str, str] = None,
            limits: Dict[str, str] = None,
            envs: Dict[str, Union[str, Secret, V1EnvVarSource]] = None,
            init_containers: Optional[List[V1alpha1UserContainer]] = None,
            sidecars: Optional[List[V1alpha1UserContainer]] = None,
            script_rendered: bool = False,
    ) -> None:
        super().__init__(name=name, inputs=inputs, outputs=outputs,
                         memoize_key=memoize_key, pvcs=pvcs,
                         annotations=annotations, labels=labels)
        self.image = image
        if isinstance(command, str):
            command = [command]
        self.command = command
        self.script = script
        if volumes is None:
            volumes = []
        self.volumes = volumes
        if mounts is None:
            mounts = []
        self.mounts = mounts
        self.init_progress = init_progress
        self.timeout = timeout
        self.retry_strategy = retry_strategy
        self.resource = resource
        self.image_pull_policy = image_pull_policy
        self.requests = requests
        self.limits = limits
        self.envs = envs
        if init_containers is None:
            init_containers = []
        self.init_containers = init_containers
        if sidecars is None:
            sidecars = []
        self.sidecars = sidecars
        self.node_selector = node_selector if node_selector is not None else {}
        self.tolerations = tolerations if tolerations is not None else []
        self.affinity = affinity
        self.script_rendered = script_rendered

    @classmethod
    def from_dict(cls, d):
        kwargs = {
            "name": d.get("name", None),
            "inputs": Inputs.from_dict(d.get("inputs", {})),
            "outputs": Outputs.from_dict(d.get("outputs", {})),
            "memoize_key": d.get("memoize", {}).get("key", None),
            "annotations": d.get("metadata", {}).get("annotations", None),
            "image": d.get("script", {}).get("image", None),
            "command": d.get("script", {}).get("command", None),
            "script": d.get("script", {}).get("source", None),
            "volumes": d.get("volumes", None),
            "mounts": d.get("script", {}).get("volumeMounts", None),
            "init_progress": d.get("metadata", {}).get("annotations", {}).get(
                "workflows.argoproj.io/progress", "0/1"),
            "timeout": d.get("timeout", None),
            "retry_strategy": d.get("retryStrategy", None),
            "resource": d.get("resource", None),
            "image_pull_policy": d.get("script", {}).get("imagePullPolicy",
                                                         None),
            "requests": d.get("script", {}).get("resources", {}).get(
                "requests", None),
            "limits": d.get("script", {}).get("resources", {}).get(
                "limits", None),
            "envs": {env["name"]: env["value"] for env in d.get(
                "script", {}).get("env", [])},
            "init_containers": d.get("initContainers", None),
            "sidecars": d.get("sidecars", None),
            "node_selector": d.get("nodeSelector", None),
            "tolerations": d.get("tolerations", None),
            "affinity": d.get("affinity", None),
        }
        engine = d.get("metadata", {}).get("annotations", {}).get(
            "workflow.dp.tech/container_engine")
        docker = "docker" if engine == "docker" else None
        singularity = "singularity" if engine == "singularity" else None
        podman = "podman" if engine == "podman" else None
        if kwargs["annotations"].get("workflow.dp.tech/script_rendered") == \
                "true":
            kwargs["script_rendered"] = True
        if d.get("metadata", {}).get("annotations", {}).get(
                "workflow.dp.tech/executor") == "dispatcher":
            host = kwargs["annotations"].get("workflow.dp.tech/host")
            port = kwargs["annotations"].get("workflow.dp.tech/port", 22)
            username = kwargs["annotations"].get(
                "workflow.dp.tech/username", "root")
            password = kwargs["annotations"].get("workflow.dp.tech/password")
            queue_name = kwargs["annotations"].get(
                "workflow.dp.tech/queue_name")
            extras = kwargs["annotations"].get("workflow.dp.tech/extras")
            extras = json.loads(extras) if extras else {}
            machine = extras.get("machine", None)
            resources = extras.get("resources", None)
            task = extras.get("task", None)
            clean = extras.get("clean", True)
            from .plugins.dispatcher import DispatcherExecutor
            executor = DispatcherExecutor(
                host, queue_name, port, username, password,
                machine_dict=machine, resources_dict=resources,
                task_dict=task, docker_executable=docker,
                singularity_executable=singularity, podman_executable=podman,
                clean=clean)
            return executor.render(cls(**kwargs))
        elif engine:
            from .executor import ContainerExecutor
            executor = ContainerExecutor(docker, singularity, podman)
            return executor.render(cls(**kwargs))
        return cls(**kwargs)

    def convert_to_argo(self, memoize_prefix=None,
                        memoize_configmap="dflow"):
        self.handle_key(memoize_prefix, memoize_configmap)
        self.annotations["workflows.argoproj.io/progress"] = self.init_progress
        if self.script_rendered:
            self.annotations["workflow.dp.tech/script_rendered"] = "true"
        if self.resource is not None:
            return V1alpha1Template(name=self.name,
                                    metadata=V1alpha1Metadata(
                                        annotations=self.annotations),
                                    inputs=self.inputs.convert_to_argo(),
                                    outputs=self.outputs.convert_to_argo(),
                                    timeout=self.timeout,
                                    retry_strategy=self.retry_strategy,
                                    memoize=self.memoize,
                                    volumes=self.volumes,
                                    resource=self.resource,
                                    init_containers=self.init_containers,
                                    sidecars=self.sidecars,
                                    )
        else:
            if self.envs is not None:
                env = []
                for k, v in self.envs.items():
                    if isinstance(v, V1EnvVarSource):
                        env.append(V1EnvVar(name=k, value_from=v))
                    elif isinstance(v, Secret):
                        env.append(V1EnvVar(name=k, value_from=V1EnvVarSource(
                            secret_key_ref=V1SecretKeySelector(
                                name=v.secret_name, key=v.secret_key))))
                    else:
                        env.append(V1EnvVar(name=k, value=v))
            else:
                env = None
            if s3_config["prefix"] and s3_config["repo"] is not None:
                loc = deepcopy(s3_config["repo"])
                loc[s3_config["repo_type"]]["key"] = \
                    "%s%s{{workflow.name}}/{{pod.name}}" % (
                        s3_config["repo_prefix"], s3_config["prefix"])
            else:
                loc = None
            return \
                V1alpha1Template(name=self.name,
                                 metadata=V1alpha1Metadata(
                                     annotations=self.annotations,
                                     labels=self.labels),
                                 node_selector=self.node_selector,
                                 tolerations=self.tolerations,
                                 affinity=self.affinity,
                                 inputs=self.inputs.convert_to_argo(),
                                 outputs=self.outputs.convert_to_argo(),
                                 timeout=self.timeout,
                                 retry_strategy=self.retry_strategy,
                                 memoize=self.memoize,
                                 volumes=self.volumes,
                                 script=V1alpha1ScriptTemplate(
                                     image=self.image,
                                     image_pull_policy=self.image_pull_policy,
                                     command=self.command, source=self.script,
                                     volume_mounts=self.mounts,
                                     resources=V1ResourceRequirements(
                                         limits=self.limits,
                                         requests=self.requests),
                                     env=env),
                                 init_containers=self.init_containers,
                                 sidecars=self.sidecars,
                                 archive_location=loc,
                                 )

    def convert_to_graph(self):
        return {
            "name": self.name,
            "type": "ScriptOPTemplate",
            "inputs": self.inputs.convert_to_graph(),
            "outputs": self.outputs.convert_to_graph(),
            "pvcs": self.pvcs,
            "annotations": self.annotations,
            "labels": self.labels,
            "image": self.image,
            "command": self.command,
            "script": self.script,
            "volumes": self.volumes,
            "mounts": self.mounts,
            "init_progress": self.init_progress,
            "timeout": self.timeout,
            "retry_strategy": self.retry_strategy,
            "resource": self.resource,
            "image_pull_policy": self.image_pull_policy,
            "requests": self.requests,
            "limits": self.limits,
            "envs": self.envs,
            "init_containers": self.init_containers,
            "sidecars": self.sidecars,
            "node_selector": self.node_selector,
            "tolerations": self.tolerations,
            "affinity": self.affinity,
        }

    @classmethod
    def from_graph(cls, graph):
        assert graph.pop("type") == "ScriptOPTemplate"
        graph["inputs"] = Inputs.from_graph(graph.get("inputs", {}))
        graph["outputs"] = Outputs.from_graph(graph.get("outputs", {}))
        return cls(**graph)


class ShellOPTemplate(ScriptOPTemplate):
    """
    Shell script OP template

    Args:
        name: the name of the OP template
        inputs: input parameters and input artifacts
        outputs: output parameters and output artifacts
        image: image the template uses
        command: command to run the script
        script: shell script
        volumes: volumes the template uses
        mounts: volumes the template mounts
        init_progress: a str representing the initial progress
        timeout: timeout of the OP template
        retry_strategy: retry strategy of the OP template
        memoize_key: memoized key of the OP template
        pvcs: PVCs need to be declared
        image_pull_policy: Always, IfNotPresent, Never
        annotations: annotations for the OP template
        labels: labels for the OP template
        node_selector: node selector when scheduling the pod
        tolerations: tolerations of taints when scheduling the pod
        affinity: affinity when scheduling the pod
        requests: a dict of resource requests
        limits: a dict of resource limits
        envs: environment variables
        init_containers: init containers before the template runs
        sidecars: sidecar containers
    """

    def __init__(
        self,
        name: Optional[str] = None,
        inputs: Optional[Inputs] = None,
        outputs: Optional[Outputs] = None,
        memoize_key: Optional[str] = None,
        pvcs: Optional[List[PVC]] = None,
        annotations: Dict[str, str] = None,
        labels: Dict[str, str] = None,
        node_selector: Dict[str, str] = None,
        tolerations: List[V1Toleration] = None,
        affinity: V1Affinity = None,
        image: Optional[str] = None,
        command: Union[str, List[str]] = None,
        script: Optional[str] = None,
        volumes: Optional[List[V1Volume]] = None,
        mounts: Optional[List[V1VolumeMount]] = None,
        init_progress: str = "0/1",
        timeout: Optional[str] = None,
        retry_strategy: Optional[V1alpha1RetryStrategy] = None,
        image_pull_policy: Optional[str] = None,
        requests: Dict[str, str] = None,
        limits: Dict[str, str] = None,
        envs: Dict[str, str] = None,
        init_containers: Optional[List[V1alpha1UserContainer]] = None,
        sidecars: Optional[List[V1alpha1UserContainer]] = None,
        script_rendered: bool = False,
    ) -> None:
        if command is None:
            command = ["sh"]
        super().__init__(
            name=name, inputs=inputs, outputs=outputs, image=image,
            command=command, script=script, volumes=volumes, mounts=mounts,
            init_progress=init_progress, timeout=timeout,
            retry_strategy=retry_strategy, memoize_key=memoize_key, pvcs=pvcs,
            image_pull_policy=image_pull_policy, annotations=annotations,
            labels=labels, node_selector=node_selector, affinity=affinity,
            tolerations=tolerations, requests=requests, limits=limits,
            envs=envs, init_containers=init_containers, sidecars=sidecars,
            script_rendered=script_rendered,
        )


class PythonScriptOPTemplate(ScriptOPTemplate):
    """
    Python script OP template

    Args:
        name: the name of the OP template
        inputs: input parameters and input artifacts
        outputs: output parameters and output artifacts
        image: image the template uses
        command: command to run the script
        script: python script
        volumes: volumes the template uses
        mounts: volumes the template mounts
        init_progress: a str representing the initial progress
        timeout: timeout of the OP template
        retry_strategy: retry strategy of the OP template
        memoize_key: memoized key of the OP template
        pvcs: PVCs need to be declared
        image_pull_policy: Always, IfNotPresent, Never
        annotations: annotations for the OP template
        labels: labels for the OP template
        node_selector: node selector when scheduling the pod
        tolerations: tolerations of taints when scheduling the pod
        affinity: affinity when scheduling the pod
        requests: a dict of resource requests
        limits: a dict of resource limits
        envs: environment variables
        init_containers: init containers before the template runs
        sidecars: sidecar containers
    """

    def __init__(
        self,
        name: Optional[str] = None,
        inputs: Optional[Inputs] = None,
        outputs: Optional[Outputs] = None,
        memoize_key: Optional[str] = None,
        pvcs: Optional[List[PVC]] = None,
        annotations: Dict[str, str] = None,
        labels: Dict[str, str] = None,
        node_selector: Dict[str, str] = None,
        tolerations: List[V1Toleration] = None,
        affinity: V1Affinity = None,
        image: Optional[str] = None,
        command: Union[str, List[str]] = None,
        script: Optional[str] = None,
        volumes: Optional[List[V1Volume]] = None,
        mounts: Optional[List[V1VolumeMount]] = None,
        init_progress: str = "0/1",
        timeout: Optional[str] = None,
        retry_strategy: Optional[V1alpha1RetryStrategy] = None,
        image_pull_policy: Optional[str] = None,
        requests: Dict[str, str] = None,
        limits: Dict[str, str] = None,
        envs: Dict[str, str] = None,
        init_containers: Optional[List[V1alpha1UserContainer]] = None,
        sidecars: Optional[List[V1alpha1UserContainer]] = None,
        script_rendered: bool = False,
    ) -> None:
        if command is None:
            command = ["python3"]
        super().__init__(
            name=name, inputs=inputs, outputs=outputs, image=image,
            command=command, script=script, volumes=volumes, mounts=mounts,
            init_progress=init_progress, timeout=timeout,
            retry_strategy=retry_strategy, memoize_key=memoize_key, pvcs=pvcs,
            image_pull_policy=image_pull_policy, annotations=annotations,
            labels=labels, node_selector=node_selector, affinity=affinity,
            tolerations=tolerations, requests=requests, limits=limits,
            envs=envs, init_containers=init_containers, sidecars=sidecars,
            script_rendered=script_rendered,
        )


class ContainerOPTemplate(ScriptOPTemplate):
    def __init__(
            self,
            command: Union[str, List[str]] = "",
            args: List[str] = None,
            **kwargs):
        if args is None:
            args = []
        if isinstance(command, str):
            command = [command]
        script = "%s %s" % (" ".join(command), " ".join(args))
        kwargs["command"] = ["sh"]
        kwargs["script"] = script
        super().__init__(**kwargs)

    @classmethod
    def from_dict(cls, d):
        kwargs = {
            "name": d.get("name", None),
            "inputs": Inputs.from_dict(d.get("inputs", {})),
            "outputs": Outputs.from_dict(d.get("outputs", {})),
            "memoize_key": d.get("memoize", {}).get("key", None),
            "annotations": d.get("metadata", {}).get("annotations", None),
            "image": d.get("container", {}).get("image", None),
            "command": d.get("container", {})["command"],
            "args": d.get("container", {}).get("args", []),
            "volumes": d.get("volumes", None),
            "mounts": d.get("container", {}).get("volumeMounts", None),
            "init_progress": d.get("metadata", {}).get("annotations", {}).get(
                "workflows.argoproj.io/progress", "0/1"),
            "timeout": d.get("timeout", None),
            "retry_strategy": d.get("retryStrategy", None),
            "resource": d.get("resource", None),
            "image_pull_policy": d.get("container", {}).get("imagePullPolicy",
                                                            None),
            "requests": d.get("container", {}).get("resources", {}).get(
                "requests", None),
            "limits": d.get("container", {}).get("resources", {}).get(
                "limits", None),
            "envs": {env["name"]: env["value"] for env in d.get(
                "container", {}).get("env", [])},
            "init_containers": d.get("initContainers", None),
            "sidecars": d.get("sidecars", None),
        }
        return cls(**kwargs)



import logging
import os
import sys
from copy import deepcopy
from typing import Dict, List, Optional, Union

from .common import (input_artifact_pattern, input_parameter_pattern,
                     task_output_artifact_pattern,
                     task_output_parameter_pattern)
from .config import config, s3_config
from .context_syntax import GLOBAL_CONTEXT
from .io import Inputs, Outputs
from .op_template import OPTemplate
from .step import add_slices
from .task import Task

try:
    from argo.workflows.client import (V1alpha1DAGTemplate, V1alpha1Metadata,
                                       V1alpha1Template)
except Exception:
    pass


class DAG(OPTemplate):
    """
    DAG

    Args:
        name: the name of the dag
        inputs: inputs in the template
        outputs: outputs in the template
        tasks: a list of tasks
        memoize_key: memoized key of the dag
        annotations: annotations for the OP template
        parallelism: maximum number of running pods for the OP template
        """

    def __init__(
            self,
            name: Optional[str] = None,
            inputs: Optional[Inputs] = None,
            outputs: Optional[Outputs] = None,
            tasks: Optional[List[Task]] = None,
            memoize_key: Optional[str] = None,
            annotations: Dict[str, str] = None,
            parallelism: Optional[int] = None,
    ) -> None:
        super().__init__(name=name, inputs=inputs, outputs=outputs,
                         memoize_key=memoize_key, annotations=annotations)
        self.parallelism = parallelism
        self.tasks = []
        if tasks is not None:
            for task in tasks:
                self.add(task)

    @classmethod
    def from_dict(cls, d, templates):
        kwargs = {
            "name": d.get("name", None),
            "inputs": Inputs.from_dict(d.get("inputs", {})),
            "outputs": Outputs.from_dict(d.get("outputs", {})),
            "memoize_key": d.get("memoize", {}).get("key", None),
            "annotations": d.get("metadata", {}).get("annotations", None),
            "parallelism": d.get("parallelism", None),
        }
        tasks = {task["name"]: Task.from_dict(task, templates)
                 for task in d.get("dag", {}).get("tasks", [])}
        for task in tasks.values():
            task.dependencies = [tasks[t] for t in task.dependencies]
        kwargs["tasks"] = list(tasks.values())
        return cls(**kwargs)

    @classmethod
    def from_graph(cls, graph, templates):
        assert graph.pop("type") == "DAG"
        graph["inputs"] = Inputs.from_graph(graph.get("inputs", {}))
        tasks = graph.pop("tasks")
        outputs = graph.pop("outputs", {})

        obj = cls(**graph)
        templates[graph["name"]] = obj

        task_dict = {}
        for t in tasks:
            task = Task.from_graph(t, templates)
            obj.add(task)
            task_dict[task.name] = task

        # replace variable references with pointers
        for task in obj.tasks:
            for name, par in task.inputs.parameters.items():
                value = getattr(par, "value", None)
                if isinstance(value, str):
                    match = input_parameter_pattern.match(value)
                    if match:
                        task.set_parameters({name: obj.inputs.parameters[
                            match.group(1)]})
                    match = task_output_parameter_pattern.match(value)
                    if match:
                        task.set_parameters({name: task_dict[match.group(
                            1)].outputs.parameters[match.group(2)]})
            for name, art in task.inputs.artifacts.items():
                source = art.source
                if isinstance(source, str):
                    match = input_artifact_pattern.match(source)
                    if match:
                        task.set_artifacts({name: obj.inputs.artifacts[
                            match.group(1)]})
                    match = task_output_artifact_pattern.match(source)
                    if match:
                        task.set_artifacts({name: task_dict[match.group(
                            1)].outputs.artifacts[match.group(2)]})

        obj.outputs = Outputs.from_graph(outputs)
        # replace variable references with pointers
        for par in obj.outputs.parameters.values():
            _from = par.value_from_parameter
            if isinstance(_from, str):
                match = input_parameter_pattern.match(_from)
                if match:
                    par.value_from_parameter = obj.inputs.parameters[
                        match.group(1)]
                match = task_output_parameter_pattern.match(_from)
                if match:
                    par.value_from_parameter = task_dict[match.group(1)].\
                        outputs.parameters[match.group(2)]
        for art in obj.outputs.artifacts.values():
            _from = art._from
            if isinstance(_from, str):
                match = input_artifact_pattern.match(_from)
                if match:
                    art._from = obj.inputs.artifacts[match.group(1)]
                match = task_output_artifact_pattern.match(_from)
                if match:
                    art._from = task_dict[match.group(1)].outputs.\
                        artifacts[match.group(2)]
        return obj

    def __iter__(self):
        return iter(self.tasks)

    def add(
            self,
            task: Union[Task, List[Task]],
    ) -> None:
        """
        Add a task or a list of tasks to the dag

        Args:
            task: a task or a list of tasks to be added to the dag
        """
        if not isinstance(task, list):
            task = [task]

        for t in task:
            assert isinstance(t, Task)
            if t.prepare_step is not None:
                self.tasks.append(t.prepare_step)
            if t.parallelism is not None:
                if self.parallelism is None:
                    self.parallelism = t.parallelism
                else:
                    self.parallelism = min(self.parallelism, t.parallelism)
                    logging.warn("Multiple tasks specified parallelism, small"
                                 "er one %s will be used" % self.parallelism)
            self.tasks.append(t)
            if t.check_step is not None:
                self.tasks.append(t.check_step)

    def convert_to_argo(self, memoize_prefix=None,
                        memoize_configmap="dflow", context=None):
        argo_tasks = []
        templates = []
        assert len(self.tasks) > 0, "Dag %s is empty" % self.name
        for task in self.tasks:
            argo_tasks.append(task.convert_to_argo(context))
            templates.append(task.template)
            templates += [hook.template for hook in task.hooks.values()]

        self.handle_key(memoize_prefix, memoize_configmap)
        argo_template = \
            V1alpha1Template(name=self.name,
                             metadata=V1alpha1Metadata(
                                 annotations=self.annotations),
                             dag=V1alpha1DAGTemplate(
                                 tasks=argo_tasks,
                             ),
                             inputs=self.inputs.convert_to_argo(),
                             outputs=self.outputs.convert_to_argo(),
                             memoize=self.memoize,
                             parallelism=self.parallelism)
        return argo_template, templates

    def convert_to_graph(self):
        graph_tasks = []
        templates = []
        for task in self.tasks:
            if not task.name.endswith("-init-artifact") and \
                not task.name.endswith("-check-num-success") and \
                    not task.name.endswith("-check-success-ratio"):
                graph_tasks.append(task.convert_to_graph())
                templates.append(task.template)

        return {
            "type": "DAG",
            "name": self.name,
            "annotations": self.annotations,
            "tasks": graph_tasks,
            "inputs": self.inputs.convert_to_graph(),
            "outputs": self.outputs.convert_to_graph(),
            "parallelism": self.parallelism,
        }, templates

    def resolve(self, pool, futures):
        import concurrent.futures
        self_copy = deepcopy(self)
        for task in self.waiting.copy():
            ready = True
            for dep in task.dependencies:
                if dep in self.finished:
                    if getattr(dep, "phase", None) != "Succeeded":
                        self.waiting.remove(task)
                        self.finished.append(task)
                        ready = False
                        break
                else:
                    ready = False
                    break
            if ready:
                task.phase = "Pending"
                i = self.tasks.index(task)
                try:
                    future = pool.submit(
                        task.run_with_config, self_copy, self.context, config,
                        s3_config, self.cwd, str(i))
                except concurrent.futures.process.BrokenProcessPool as e:
                    # retrieve exception of subprocess before exit
                    for future in concurrent.futures.as_completed(futures):
                        future.result()
                    raise e
                futures[future] = i
                self.waiting.remove(task)
                self.running.append(task)

    def run(self, workflow_id=None, context=None, stepdir=None):
        self.workflow_id = workflow_id
        self.stepdir = stepdir
        self.context = context
        import concurrent.futures
        self.cwd = os.getcwd()
        max_workers = config["debug_pool_workers"]
        if max_workers == -1:
            max_workers = len(self.tasks)
        if max_workers is None:
            max_workers = os.cpu_count() or 1
        max_workers = min(max_workers, len(self.tasks)) or 1
        pool = concurrent.futures.ProcessPoolExecutor(max_workers)
        futures = {}
        self.waiting = [task for task in self]
        self.running = []
        self.finished = []
        self.resolve(pool, futures)

        while len(self.running) > 0:
            future = next(concurrent.futures.as_completed(futures))
            j = futures.pop(future)
            try:
                phase, pars, arts = future.result()
            except Exception:
                import traceback
                traceback.print_exc()
                self.tasks[j].phase = "Failed"
                if not self.tasks[j].continue_on_failed:
                    if sys.version_info.minor >= 9:
                        pool.shutdown(wait=False)
                    else:
                        pool.shutdown(wait=True)
                    raise RuntimeError("Task %s failed" % self.tasks[j])
            else:
                for name, value in pars.items():
                    self.tasks[j].outputs.parameters[
                        name].value = value
                for name, path in arts.items():
                    self.tasks[j].outputs.artifacts[
                        name].local_path = path
                self.tasks[j].phase = phase
                logging.info("Outputs of %s collected" % self.tasks[j])
            self.running.remove(self.tasks[j])
            self.finished.append(self.tasks[j])
            self.resolve(pool, futures)

        if sys.version_info.minor >= 9:
            pool.shutdown(wait=False)
        else:
            pool.shutdown(wait=True)
        assert len(self.finished) == len(self.tasks), "cyclic graph"

    def add_slices(self, slices, layer=0):
        add_slices(self, slices, layer=layer)

    def copy(self):
        new_template = self.deepcopy()
        for task, new_task in zip(self.tasks, new_template.tasks):
            new_task.template = task.template
        return new_template

    def __enter__(self) -> 'DAG':
        GLOBAL_CONTEXT.in_context = True
        GLOBAL_CONTEXT.current_workflow = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        GLOBAL_CONTEXT.in_context = False
